# REPRAP GCODE to contours (positions and speed (extrusion and movement))
# reference:
# https://reprap.org/wiki/G-code

import re
import numpy as np
from math import ceil
# import matplotlib.pyplot as plt
# import matplotlib.colors as mcolors
import networkx as nx
from copy import copy
# import plotly.graph_objects as go


class Position(tuple):
    """
    Wrapper for Tuples makes position operations easier
    - difference between points
    - distance between points

    Example: getting the distance between two postions
        > abs(pos1 - pos2)
    """

    def __new__(self, x, y, z):
        return tuple.__new__(Position, (x,y,z))

    def __sub__(self, pos2):
        return tuple.__new__(Position, (self[0]-pos2[0], self[1]-pos2[1], self[2]-pos2[2]))

    def __add__(self, pos2):
        return tuple.__new__(Position, (self[0]+pos2[0], self[1]+pos2[1], self[2]+pos2[2]))

    def __abs__(self):
        return (self[0]**2 + self[1]**2 + self[2]**2)**0.5

    def __getnewargs__(self):
        return (self[0], self[1], self[2])


class Contour(object):
    """
    Contour Object
    Stores 1 Contour including following information:
        - Waypoint positions
        - Waypoint Extrusion position (relative to start)
        - Waypoint velocity
        - Dependencies (contours which it depends on)
    """

    def __init__(self, pos_list, start_extrusion=0, normals=None, start_velocity=0, comment=None):

        # List of physical positions of waypoints
        self.pos = pos_list
        self.time = [0]
        self.start_ext = start_extrusion
        self.Ext = [start_extrusion]
        self.Vel = [start_velocity]
        self.id = None
        normals = normals
        self.comment = "" if comment is None else comment

    # Get the nth waypoint in the contour
    def __getitem__(self, n):
        return self.pos[n]

    # Number of waypoints
    def __len__(self):
        return len(self.pos)

    # TODO Total length of contour
    def dist(self):
        pass


def split_contour_waypoints(c, max_wp_dist):
    """
    add additional waypoints where two are too wide apart
    """
    i=1
    new_pos = []
    new_time = []
    new_Ext = []
    while i<len(c):
        d = abs(c.pos[i-1] - c.pos[i])
        if d > max_wp_dist:
            ps = list(np.linspace(c.pos[i-1], c.pos[i], ceil(d/max_wp_dist)+1))
            ts = np.linspace(c.time[i-1], c.time[i], ceil(d/max_wp_dist)+1)
            Es = np.linspace(c.Ext[i-1], c.Ext[i], ceil(d/max_wp_dist)+1)
            for j in range(int(d/max_wp_dist)+1):
                new_pos.append(Position(*tuple(ps[j])))
                new_time.append(ts[j])
                new_Ext.append(Es[j])
        else:
            new_pos.append(c.pos[i-1])
            new_time.append(c.time[i-1])
            new_Ext.append(c.Ext[i-1])
        i+=1

    new_pos.append(c.pos[-1])
    new_time.append(c.time[-1])
    new_Ext.append(c.Ext[-1])

    c.pos = new_pos
    c.time = new_time
    c.Ext = new_Ext
    return c


class machine:
    """
    Machine State
    Stores the state of the machine given GCode lines so far.

    Parses line of Gcode from String and updates its state to whate it would be
        after following that line of GCode instruction
    """

    def __init__(self):

        self.x = 0
        self.y = 0
        self.z = 0
        self.Ext = 0
        self.Vel = 0
        self.units = "mm"
        self.tempExt = 0
        self.tempBed = 0
        self.fan = False
        self.extruding = False
        self.last_state = None

    def read_gcode_line(self, l):
        """
        Supported GCODE lines:
        M107 - Fan off
        M106 - Fan on
        M104 - Set Extruder Temperature
        G28  - Move to origin
        G1   - Linear motion
        G92  - Set Position (current position, new coordinates)
            WARNING: Assumed G92 is always used to reset Extruder origind

        TODO:
        M84  - Idle hold
        M109 - Wait for temperature to be reached
        M82  - use absolute distances for extrusion
        G90  - use absolute coordinates
        """

        self.last_state = self.get_state()

        if l.startswith('G1 ') or l.startswith('G0 '):
            # Strip Comments
            if ';' in l:
                l, comment = l.split(';')

            # Position Updates
            if 'X' in l:
                self.x = float(re.search(r"X(-?\d+.?(?:\d+)?)", l).group(1))
            if 'Y' in l:
                self.y = float(re.search(r"Y(-?\d+.?(?:\d+)?)", l).group(1))
            if 'Z' in l:
                self.z = float(re.search(r"Z(-?\d+.?(?:\d+)?)", l).group(1))

            # Velocity Update
            if 'F' in l:
                self.Vel = float(re.search(r"F(-?\d+.?(?:\d+)?)", l).group(1))

            # Extrusion Updates
            if 'E' in l:
                self.Ext = float(re.search(r"E(-?\d+.?(?:\d+)?)", l).group(1))
                self.extruding = True
                # if self.Ext == 0:
                #     self.extruding = False
            else:
                self.extruding = False

        elif l.startswith('G92'):
            self.Ext = 0
            self.extruding = False
        elif l.startswith(';') or l.startswith("\n"):
            return False
        elif l.startswith('M107'):
            self.fan = False
        elif l.startswith('M106'):
            self.fan = True
        elif l.startswith('M104'):
            self.tempExt = re.search(r"S(\d+)", l).group(1)
        elif l.startswith('G28'):
            l = l.split(';')[0]
            if 'X' in l:
                self.x = 0
            if 'Y' in l:
                self.y = 0
            if 'Z' in l:
                self.z = 0
            if not any(z in l for z in ['X', 'Y', 'Z']):
                self.x = 0
                self.y = 0
                self.z = 0
        elif l.startswith('M84'):
            print("NOT IMPLEMENTED: Idle hold")
        elif l.startswith('M109'):
            print("NOT IMPLEMENTED: Wait for temperature to be reached")
        elif l.startswith('M82'):
            print("NOT IMPLEMENTED: use absolute distances for extrusion")
        elif l.startswith('G90'):
            print("NOT IMPLEMENTED: use absolute coordinates")
        else:
            print("ERROR: unknown GCODE command " + l.split(' ')[0])

    def get_state(self):
        state = {
                'x': self.x,
                'y': self.y,
                'z': self.z,
                'Vel': self.Vel,
                'Ext': self.Ext,
                'fan': self.fan,
                'tempExt': self.tempExt,
                'tempBed': self.tempBed,
                'extruding': self.extruding
                }
        return state


class DependencyGraph:
    """
    Holds the dependency graph for a set of contours and the current
        state of what is available given which contours have been done

    NOTE: Edges in the graph are directed dependency -> dependent
    """

    def __init__(self, contours, dep_range, max_layers, c_layers=None):

        # save inputs for reference
        self.all_contours = contours
        self.dep_range = dep_range
        self.max_layers = max_layers

        # create the graph (nodes, no edges)
        self.G = nx.DiGraph()
        if c_layers is None:
            # figure out layers (each layer a list of contours)
            layer_heights = list(set([c.pos[0][2] for c in contours]))
            layer_heights.sort()
            self.layers = [np.array([], dtype=int) for i in layer_heights]

            for c in contours:
                l = np.searchsorted(layer_heights,c.pos[0][2])
                self.G.add_node(c.id, contour=c, layer=l) #node for each contour
                self.layers[l] = np.append(self.layers[l], c.id)
        else:
            c_layers = np.array(c_layers)
            self.layers = [np.where(c_layers==i)[0] for i in range(max(c_layers)+1)]
            for l_i,l in enumerate(self.layers):
                for c in l:
                    self.G.add_node(c, contour=contours[c], layer=l_i)

        # add edges to graph showing dependencies
        for n in self.G.nodes:
            l = self.G.nodes[n]['layer']
            # Add every contour from max_layers below to dependencies
            if l >= max_layers:
                for c in self.layers[l - max_layers]:
                    self.G.add_edge(c,n)

            # Add dependencies from layer below
            if l == 0 or max_layers == 1:
                continue
            for c in self.layers[l-1]:
                # If any waypoint is closer than the specified distance
                dep_found = False
                for i,a in enumerate(contours[c].pos):
                    if i%2==0:
                        continue
                    for j,b in enumerate(contours[n].pos):
                        if j%2==0:
                            continue
                        if abs(a-b) < dep_range:
                            self.G.add_edge(c,n)
                            dep_found = True
                            break
                    if dep_found:
                        break

        # set variables to the starting state
        self.reset_state()


    def reset_state(self):
        """
        Remove all done contours
        """
        self.done_contour_ids = []
        self.avail_contour_ids = self.layers[0].tolist()
        self.available_contours = [self.G.nodes[n]['contour']
                for n in self.avail_contour_ids]
        self.todo_contour_ids = [c.id for c in self.all_contours]
        self.done_contour_ids_set = set(self.done_contour_ids)
        self.avail_contour_ids_set = set(self.avail_contour_ids)
        self.todo_contour_ids_set = set(self.todo_contour_ids)
        return


    def do_contour(self, contour_id, validate=True, save_available=True):
        """
        Re-calculate available contours based on new contour completion
        """

        if validate:
            # make sure that the contour dependencies have been satisfied
            assert isinstance(contour_id, (int, np.integer)), "Use contour_id, not contour object"
            if contour_id not in self.avail_contour_ids:
                if contour_id in self.done_contour_ids:
                    raise Exception("This contour has already been done")
                elif not all([d in self.done_contour_ids for d,_ in self.G.in_edges(contour_id)]):
                    unsat_deps = []
                    for d,_ in self.G.in_edges(contour_id):
                        if d not in self.done_contour_ids:
                            unsat_deps.append(d)
                    raise Exception("Contour dependencies not yet Satisfied:" + str(unsat_deps))
                else:
                    raise Exception("Something's gone wrong here")

        # Move from available to done
        self.done_contour_ids.append(contour_id)
        self.avail_contour_ids.remove(contour_id)
        self.todo_contour_ids.remove(contour_id)

        # Recalculate what is available
        for c in self.todo_contour_ids:

            if c in self.avail_contour_ids:
                continue

            # for the case with no dependencies
            if len(self.G.in_edges(c)) == 0:
                self.avail_contour_ids.append(c)
                continue

            # others
            dependencies_satisfied = True
            for d,_ in self.G.in_edges(c):
                if d not in self.done_contour_ids:
                    dependencies_satisfied = False
                    break
            if dependencies_satisfied:
                self.avail_contour_ids.append(c)
        
        if save_available:
            # save available contours
            self.available_contours = [self.G.nodes[n]['contour']
                    for n in self.avail_contour_ids]
        return


    def do_contour2(self, contour_id):
        """
        Refactored version of do_contour with a lower time complexity
        """
        # Move from available to done
        self.done_contour_ids_set.add(contour_id)
        self.avail_contour_ids_set.discard(contour_id)
        self.todo_contour_ids_set.discard(contour_id)

        # Get the set of all contours with no dependencies
        no_dependency_contours = {c for c in self.todo_contour_ids_set if len(self.G.in_edges(c)) == 0}
        
        # Get the set of contours with all dependencies satisfied
        satisfied_dependencies = {c for c in self.todo_contour_ids_set if all(d in self.done_contour_ids_set for d, _ in self.G.in_edges(c))}
        
        # Preserve the order while updating avail_contour_ids
        updated_avail_contours = [c for c in self.todo_contour_ids if c in no_dependency_contours or c in satisfied_dependencies]

        # Convert updated_avail_contours to a set for faster membership testing
        self.avail_contour_ids = set(updated_avail_contours)

        return


    def __getitem__(self, contour_id):
        return self.G.nodes[contour_id]['contour']


def decode_gcode2(fname):
    """
    new version of decode gcode
    """
    printer = machine()
    contours = []
    c = None
    with open(fname, 'r') as f:
        for line in f.readlines():
            printer.read_gcode_line(line)
            if not printer.extruding:
                if c is not None:
                    contours.append(c)
                c = Contour([Position(printer.x, printer.y, printer.z)],
                            start_extrusion=printer.Ext,
                            start_velocity=printer.Vel,
                            )
            else:
                if c is not None and not np.allclose([printer.last_state['x'], printer.last_state['y'], printer.last_state['z']],
                                                     [printer.x, printer.y, printer.z]):
                    c.pos.append(Position(printer.x, printer.y, printer.z))
                    c.Ext.append(printer.Ext)
                    c.Vel.append(printer.Vel)

    contours = [c for c in contours if len(c.pos) > 1]
    return contours


def decode_gcode(fname, max_wp_dist=20, max_extrusion=200, trim=True):
    """
    Convert GCODE file into a list of contours

    Input:
     - file name (String)
     - trim (bool, optional): Whether to remove non-extrusion contours and contours that don't have any movement

    Ouptut:
     - List of contour objects
    """

    printer = machine()
    c = None
    contours = []
    with open(fname, 'r') as f:

        for line in f.readlines():

            # for each line of gcode, interpret the line and then ...
            printer.read_gcode_line(line)

            # save what the state of the machine is after executing the line
            if printer.extruding and not printer.last_state['extruding']:
                c = Contour([Position(printer.last_state['x'],
                            printer.last_state['y'],
                            printer.last_state['z'])],
                            printer.last_state['Ext'],
                            )

            elif printer.extruding:
                c.pos.append(Position(printer.last_state['x'],
                             printer.last_state['y'],
                             printer.last_state['z'])
                             )
                c.Ext.append(printer.last_state['Ext'] - c.start_ext)
                c.time.append(c.time[-1] + abs(c.pos[-1]-c.pos[-2])/(printer.last_state['Vel']/60))

            elif not printer.extruding and printer.last_state['extruding']:
                c.pos.append(Position(printer.last_state['x'],
                             printer.last_state['y'],
                             printer.last_state['z'])
                             )
                c.Ext.append(printer.last_state['Ext'] - c.start_ext)
                c.time.append(c.time[-1] + abs(c.pos[-1]-c.pos[-2])/(printer.last_state['Vel']/60))

                contours.append(c)
                c = None

    if trim:
        # remove non extrusion and very short contours
        contours = [c for c in contours
                    if c.Ext[-1]>0 and (len(c.pos)>2 or
                    (len(c.pos)==2 and c.pos[0]!=c.pos[1]))]

    # split for max extrusion length
    split_contours = []
    for c in contours:
        split_contour_waypoints(c, max_wp_dist)

        # if splitting required
        if c.Ext[-1] > max_extrusion:
            i = 1
            c_out = Contour([c.pos[0]],0)
            cut_extrusion = 0
            cut_time = 0
            while i <= len(c.pos):
                # if time to split
                if i==len(c.pos) or c.Ext[i] - cut_extrusion > max_extrusion:
                    cut_extrusion = c.Ext[i-1]
                    cut_time = c.time[i-1]
                    split_contours.append(c_out)
                    if i<len(c.pos):
                        new_c = Contour([c_out.pos[-1]], 0)
                        c_out = new_c
                    else:
                        break

                c_out.pos.append(c.pos[i])
                c_out.Ext.append(c.Ext[i] - cut_extrusion)
                c_out.time.append(c.time[i] - cut_time)
                i+=1
        else:
            split_contours.append(c)

    for i, c in enumerate(split_contours):
        c.id = i
    return split_contours


# def plot_contours_sideview(*args, colors=None, axis='y'):
#     """
#     Plots a list of contours in 2d (top down)
#     if contours are not on the same layer,
#       layering may not be ordered
#     """
#     if colors is None:
#         colors = [k for k in mcolors.cnames]

#     if axis == 'y':
#         for i, contours in enumerate(args):
#             for c in contours:
#                 xs = [pos[0] for pos in c.pos]
#                 ys = [pos[1] for pos in c.pos]
#                 zs = [pos[2] for pos in c.pos]
#                 plt.plot(xs, zs, color=colors[i])
#     elif axis == 'x':
#         for i, contours in enumerate(args):
#             for c in contours:
#                 xs = [pos[0] for pos in c.pos]
#                 ys = [pos[1] for pos in c.pos]
#                 zs = [pos[2] for pos in c.pos]
#                 plt.plot(ys, zs, color=colors[i])
#     else:
#         print("axis must be 'x' or 'y'.")
#         return

#     plt.axis('equal')
#     plt.show()
#     return


# def plot_contours2d(*args, colors=None):
#     """
#     Plots a list of contours in 2d (top down)
#     if contours are not on the same layer,
#       layering may not be ordered
#     """
#     if colors is None:
#         colors = [k for k in mcolors.cnames]

#     for i, contours in enumerate(args):
#         for c in contours:
#             xs = [pos[0] for pos in c.pos]
#             ys = [pos[1] for pos in c.pos]
#             zs = [pos[2] for pos in c.pos]
#             plt.plot(xs, ys, color=colors[i])
#     plt.axis('equal')
#     plt.show()
#     return


# def plot_contours3d(*args, normals=False, colors=None):
#     """
#     Plots a list of contours

#     Each input arguement is a list of contours
#       All contours within each list will be the same color
#       Contours in different lists will be different colors
#     """
#     fig = go.Figure()
#     if colors is None:
#         colors = [k for k in mcolors.cnames]

#     for i, contours in enumerate(args):
#         for c in contours:
#             xs = [pos[0] for pos in c.pos]
#             ys = [pos[1] for pos in c.pos]
#             zs = [pos[2] for pos in c.pos]
#             fig.add_trace(
#                 go.Scatter3d(
#                     x=xs,
#                     y=ys,
#                     z=zs,
#                     mode='lines',
#                     line={"color":colors[i]},
#                     )
#                 )
#             if normals:
#                 # plot quiver of c.normals
#                 fig.add_trace(go.Cone(x=xs,y=ys,z=zs,u=c.normals[:,0],v=c.normals[:,1],w=c.normals[:,2]))
#     fig.update_layout(scene_aspectmode='data')
#     fig.show()
#     return


def split_contours(contours, max_time, zeroed=True):
    """
    Splits contours into equal length contours such that each
    contour can be executed in a given amount of time.

    Input:
     - list of contour objects
     - maximum amount of time for each contour

    Output:
     - list of contour objects that can all be executed in
       max_time amount of time
    """

    # Our new list of smaller contours
    cs = []

    for cont in contours:
        duration = cont.time[-1] - cont.time[0] # Last minus first time for duration

        # If the duration already satisfies the time limit, then we keep it
        if duration <= max_time:
            cs.append(cont)
            continue

        # Otherwise split it
        # TODO Maybe a version that splits AT nearest waypoints would be better

        # Calculate how minimum number of parts we need
        nPart = int(np.ceil(duration/max_time).item())
        t_split = np.linspace(cont.time[0],cont.time[-1],nPart+1)

        # Query at the split points
        # (Assume cont.time is monotonically increasing)
        x_split = np.interp(t_split,cont.time,[p[0] for p in cont.pos])
        y_split = np.interp(t_split,cont.time,[p[1] for p in cont.pos])
        z_split = np.interp(t_split,cont.time,[p[2] for p in cont.pos])
        e_split = np.interp(t_split,cont.time,[e for e in cont.Ext])

        # Create the needed subcontours in a merge sort kind of way
        iCont = 1
        for iSplit in range(0,nPart):
            c = Contour([Position(x_split[iSplit],y_split[iSplit],z_split[iSplit])],cont.start_ext+e_split[iSplit])

            # Add the original waypoints until we need to add next split waypoint
            while cont.time[iCont] < t_split[iSplit+1]:
                c.pos.append(cont.pos[iCont])
                c.Ext.append(cont.Ext[iCont]-c.start_ext)
                c.time.append(cont.time[iCont]-t_split[iSplit])

                iCont += 1

            # Now add the next split waypoint
            c.pos.append(Position(x_split[iSplit+1],y_split[iSplit+1],z_split[iSplit+1]))
            c.Ext.append(e_split[iSplit+1]-e_split[iSplit])
            c.time.append(t_split[iSplit+1]-t_split[iSplit])

            # NOTE: This function does not support dependencies,
            #   it is recomended they are re-calculated
            cs.append(c)
    return cs


# if __name__ == "__main__":
#
#     import sys
#     from os import path
#
#     if path.exists(sys.argv[1]):
#         contours = decode_gcode2(sys.argv[1])
#     else:
#         print("Invalid argument - file not found")
